from itertools import chain
import logging
import sys

import click
import click_completion

from . import _fetchrepodata, _repodata
from . import flatpak_generator
from . import module_generator
from . import module_repoquery
from . import modulemd_linter
from . import modulemd_summarizer
from .config import config, runtime_config
from .sources import Source
from .util import ChoiceGlob


# click only allows hiding/deprecating commands/parameters from version 7.0 on
click_version_info = tuple(int(x) for x in click.__version__.split('.'))
if click_version_info >= (7, 0):
    hidden_arg = {'hidden': True}
    hidden_deprecated_arg = {'hidden': True, 'deprecated': True}
else:
    hidden_arg = {}
    hidden_deprecated_arg = {}


# fedmod uses click for argument parsing, but currently does its own
# standardised exception handling. This also requires handling click's standard
# exception types as described in http://click.pocoo.org/5/exceptions/
def run():
    click_completion.init()
    try:
        rc = _cli_commands.main(sys.argv[1:], standalone_mode=False,
                                default_map=config.get('options', {}))
    except (KeyboardInterrupt, click.Abort):
        print('\nInterrupted by user')
        rc = 1
    except click.ClickException as e:
        e.show()
        rc = e.exit_code
    except _fetchrepodata.MissingMetadata as e:
        print(e, file=sys.stderr)
        rc = 2
    except Exception as e:
        logging.exception(f"Unexpected exception: {e!r}")
        raise
    sys.exit(rc)


@click.group()
@click.option("--default-loglevel", "loglevel", flag_value=logging.WARN,
              default=True, **hidden_arg)
@click.option("--verbose", "-v", "loglevel", flag_value=logging.INFO,
              help="Show additional info messages")
@click.option("--debug", "loglevel", flag_value=logging.DEBUG,
              help="Show debugging messages")
@click.option("--source-type", "-S", metavar="SOURCE",
              help="Set the type of source to query",
              type=click.Choice(Source._types_to_cls))
@click.option("--dataset", "-s", metavar="RELEASE[:ARCH]",
              help="Set the release to use for package and module data",
              type=click.Choice(config.datasets))
def _cli_commands(loglevel, source_type, dataset):
    """Utilities for working with modular Fedora RPM repositories"""
    _setup_app(loglevel, source_type, dataset)


def _setup_app(loglevel=logging.WARN, source_type=None, dataset=None):
    logging.basicConfig(level=loglevel)

    if dataset:
        try:
            _repodata.set_dataset_name(dataset)
        except ValueError as e:
            raise click.ClickException(e) from e
    else:
        dataset = _repodata.dataset_name

    runtime_config['dataset'] = dataset

    if not source_type:
        source_type = 'rpmrepodata'

    runtime_config['source_type'] = source_type
    runtime_config['source'] = Source.from_type(source_type)


# Fetching metadata
@_cli_commands.command('fetch-metadata')
def fetch_metadata():
    """Fetch and cache required module and RPM metadata"""
    _fetchrepodata.download_repo_metadata(_repodata.dataset_name)


# modulemd generation
@_cli_commands.command()
@click.option("--output", "-o", metavar="FILE",
              help="Write to FILE instead of stdout.")
@click.argument("pkgs", metavar='PKGS', nargs=-1, required=True)
def rpm2module(pkgs, output):
    """Generate a draft modulemd from an RPM (or list of RPMs)"""
    mg = module_generator.ModuleGenerator(pkgs)
    mg.run(output)


# modulemd generation from local repositories
@_cli_commands.command()
@click.option("--output", "-o", 'out_file', metavar='FILE',
              help="Write to FILE instead of stdout.")
@click.option("--input", "-i", 'in_file', metavar='FILE',
              help="Start with module metadata from FILE.")
@click.argument("repopath", metavar='REPOPATH', nargs=1, required=True)
@click.argument("apipkgs", metavar='APIPKGS', nargs=-1, required=False)
def repo2module(repopath, apipkgs, out_file, in_file):
    """Generate or update a modulemd from a local repository"""
    if out_file and out_file == in_file:
        raise click.BadOptionUsage('out_file', "Input and output files can't be the same.")

    mg = module_generator.ModuleFromRepoGenerator(repopath, apipkgs, in_file=in_file)
    mg.run(out_file)


# modulemd generation for Flatpaks
@_cli_commands.command()
@click.option("--flathub", metavar="ID_OR_SEARCH_TERM",
              help="Initialize from a Flathub Flatpak.")
@click.option("--flatpak-common", "-c", is_flag=True,
              help="Use dependencies from the flatpak-common module")
@click.option("--output-modulemd", metavar="FILE",
              help="Write modulemd to FILE instead of <pkg>.yaml.")
@click.option("--output-containerspec", metavar="FILE",
              help="Write container specification to FILE"
                   " instead of container.yaml.")
@click.option("--force", "-f", is_flag=True,
              help="Overwriting existing output files")
@click.argument("pkg", metavar='PKG', required=True)
def rpm2flatpak(pkg, output_modulemd, output_containerspec, force, flathub, flatpak_common):
    """Generate modulemd from an RPM"""
    fg = flatpak_generator.FlatpakGenerator(pkg)
    fg.run(output_modulemd, output_containerspec, force=force, flathub=flathub,
           include_flatpak_common=flatpak_common)


@_cli_commands.command('flatpak-report')
@click.option("--quiet", "-q", is_flag=True,
              help="Don't print progress messages to stderr.")
@click.argument("pkgs", metavar='PKGS', nargs=-1, required=True)
def flatpak_report(pkgs, quiet):
    flatpak_generator.do_flatpak_report(pkgs, quiet=quiet)


# Checking availability of dependencies through module streams
@_cli_commands.command('resolve-deps')
@click.option("--module-dependency", "-m", multiple=True, metavar="MODULE",
              help="Module to be used as a dependency."
                   " Can be given multiple times.")
@click.option("--json", is_flag=True, default=False,
              help="Output dependencies in JSON format"
                   " with extra information.")
@click.argument("pkgs", metavar='PKGS', nargs=-1, required=True)
def resolve_deps(pkgs, module_dependency, json):
    """Report dependencies of the given RPM (or list of RPMs)"""
    rq = module_repoquery.ModuleRepoquery()
    rq.list_pkg_deps(pkgs, module_dependency, json)


# Listing modules
@_cli_commands.command('list-modules')
def list_modules():
    """List the available modules"""
    rq = module_repoquery.ModuleRepoquery()
    rq.list_modules()


# Looking for RPM duplication across modules
@_cli_commands.command('list-rpms')
@click.option("--duplicate-only", is_flag=True, default=False,
              help="Only list packages that are in more than one module.")
@click.option("--list-modules", is_flag=True,
              help="List modules for every package.")
def list_rpms(duplicate_only, list_modules):
    """List RPMs available through modules"""
    rq = module_repoquery.ModuleRepoquery()
    rq.list_modularized_pkgs(
        duplicate_only=duplicate_only,
        list_modules=list_modules,
    )


# Checking module contents
# (ncoghlan): Perhaps `module-rpms` would be more consistent?
@_cli_commands.command('module-packages')
@click.option("--full-nevra", is_flag=True, default=False,
              help="Print the full NEVRA instead of only name.")
@click.argument("module")
def module_rpms(module, full_nevra):
    """Lists RPMS in a given module"""
    rq = module_repoquery.ModuleRepoquery()
    rq.list_rpms_in_module(module, full_nevra=full_nevra)


# Checking availability of specific packages
# (ncoghlan): Perhaps `where-is-rpm` would be more consistent?
@_cli_commands.command('where-is-package')
@click.argument("pkg")
def which_module(pkg):
    """Reports which module (if any) provides the given RPM"""
    rq = module_repoquery.ModuleRepoquery()
    rq.list_modules_for_rpm(pkg)


# Checking what to build to satisfy particular binary dependencies
@_cli_commands.command('srpm-of-rpm')
@click.argument("pkg")
def which_srpm(pkg):
    """Reports which SRPM produces the given RPM"""
    rq = module_repoquery.ModuleRepoquery()
    rq.get_srpm_for_rpm(pkg)


# Map SRPM to the RPMs it generates. Useful info when building
# new modules, choosing the API or what to filter
@_cli_commands.command('rpms-from-srpm')
@click.argument("pkg")
def rpms_from_srpm(pkg):
    """Reports which RPMS are generated from the given SRPM"""
    rq = module_repoquery.ModuleRepoquery()
    rq.get_rpms_for_srpm(pkg)


# Check a modulemd YAML file for validity.
@_cli_commands.command()
@click.option("--min-level", metavar="MINLEVEL", default='info',
              type=click.Choice(modulemd_linter.LintLevel.all_levels),
              help="The minimum level of checks to be performed ({}).".format(
                   ", ".join(modulemd_linter.LintLevel.all_levels)))
@click.argument("modulemd")
def lint(modulemd, min_level):
    """Validates a given modulemd YAML file"""
    linter = modulemd_linter.ModuleMDLinter(modulemd_path=modulemd)
    return linter.lint(min_level=min_level)


# Summarize a list of modulemd files
@_cli_commands.command('summarize-modules')
@click.argument("modules", metavar='MODULES', nargs=-1, required=False)
@click.option("--file", "-f", "files", metavar="FILE", multiple=True,
              type=click.Path(exists=True),
              help="Modulemd files to check."
                   " Ignores other sources like default repositories or build"
                   " systems. Can be given multiple times.")
@click.option("--url", "-u", "urls", metavar="URL", multiple=True,
              help="Repositories to read module metadata from."
                   " Ignores other sources like default repositories or build"
                   " systems. Can be given multiple times.")
@click.option("--show-property", "-p", "add_prop_names", metavar="PROPERTY",
              multiple=True, type=ChoiceGlob(modulemd_summarizer.all_property_names))
@click.option("--tree", "-t", "as_tree", is_flag=True, default=False,
              help="Format output as a tree")
def summarize_modules(modules, files, urls, add_prop_names, as_tree):
    """Prints a summary of modules"""
    modulemd_summarizer.summarize_modules(modules, files, urls, chain(*add_prop_names), as_tree)


# Summarize a list of modulemd files (deprecated name and options)
@_cli_commands.command('summarize-module', **hidden_deprecated_arg)
@click.argument("modules", metavar='MODULES', nargs=-1, required=False)
@click.option("--add-file", "-f", metavar="FILE", multiple=True,
              type=click.Path(exists=True),
              help="Modulemd files to check."
                   " Ignores other sources like default repositories or build"
                   " systems. Can be given multiple times.")
@click.option("--add-url", "-u", metavar="URL", multiple=True,
              help="Repositories to read module metadata from."
                   " Ignores other sources like default repositories or build"
                   " systems. Can be given multiple times.")
@click.option("--tree", "-t", is_flag=True, default=False,
              help="Print output as a tree")
def summarize_module(modules, add_file, add_url, tree):
    """Prints a summary of modules"""
    modulemd_summarizer.summarize_modules(modules, add_file, add_url, tree)


# Get the modulemd contents of specified modules
@_cli_commands.command('get-modulemds')
@click.argument("modules", metavar='MODULES', nargs=-1, required=False)
@click.option("--file", "-f", "files", metavar="FILE", multiple=True,
              type=click.Path(exists=True),
              help="Modulemd files to check."
                   " Ignores other sources like default repositories or build"
                   " systems. Can be given multiple times.")
@click.option("--url", "-u", "urls", metavar="URL", multiple=True,
              help="Repositories to read module metadata from."
                   " Ignores other sources like default repositories or build"
                   " systems. Can be given multiple times.")
def get_modulemds(modules, files, urls):
    """Prints a summary of modules"""
    modulemd_summarizer.summarize_modules(modules, files, urls, action='modulemds')
