import json
import os
import re
import sys
from textwrap import dedent

import click
import gi
gi.require_version('Modulemd', '1.0')  # noqa: E402
from gi.repository import Modulemd
import requests
import yaml

from . import _depchase
from ._repodata import dataset_release_name
from .get_module_builds import get_module_builds
from .module_generator import ModuleGenerator
from .util import rpm_name_only


# Some PyYAML magic to get the output we want for container.yaml

class LiteralScalar(str):
    """String subclass that gets dumped into yaml as a scalar"""
    pass


def _represent_literal_scalar(dumper, s):
    return dumper.represent_scalar(tag=u'tag:yaml.org,2002:str',
                                   value=s,
                                   style='|')


yaml.add_representer(LiteralScalar, _represent_literal_scalar)


class NoSortMapping(dict):
    """dict subclass that dumped into yaml as a scalar without sorting keys"""
    pass


def _represent_no_sort_mapping(dumper, d):
    return yaml.MappingNode(tag='tag:yaml.org,2002:map',
                            value=[(dumper.represent_data(k),
                                    dumper.represent_data(v))
                                   for k, v in d.items()],
                            flow_style=False)


yaml.add_representer(NoSortMapping, _represent_no_sort_mapping)


def dataset_release_branch():
    release_name = dataset_release_name()
    if release_name == 'rawhide':
        return 'master'
    else:
        return release_name


def _get_runtime_packages():
    builds = get_module_builds('flatpak-runtime', dataset_release_branch())
    # Each flatpak-runtime stream should be built against a single context
    assert len(builds) == 1
    build = builds[0]

    mmd_str = build['extra']['typeinfo']['module']['modulemd_str']
    mmd = Modulemd.Module.new_from_string(mmd_str)

    return set(mmd.peek_profiles()['runtime'].peek_rpms().get())


def _get_common_packages():
    builds = get_module_builds('flatpak-common', dataset_release_branch(), include_rpms=True)
    # Each flatpak-common stream should be built against a single context
    assert len(builds) == 1
    build = builds[0]

    return {r['name'] for r in build['fedmod_rpms']}


def _load_flathub_manifest(search_term):
    response = requests.get("https://flathub.org/api/v1/apps")
    response.raise_for_status()
    apps = response.json()

    matches = []
    search_lower = search_term.lower()
    for app in apps:
        if (search_lower in app['flatpakAppId'].lower() or
                search_lower in app['name'].lower()):
            matches.append((app['flatpakAppId'], app['name']))

    if len(matches) > 1:
        max_id_len = max([len(app_id) for app_id, _ in matches])
        for app_id, name in matches:
            print(app_id + (' ' * (max_id_len - len(app_id)) + ' ' + name))
        raise click.ClickException("Multiple matches found on flathub.org")
    elif len(matches) == 0:
        raise click.ClickException("No match found on flathub.org")

    app_id = matches[0][0]

    for fname, is_yaml in [
            (f"{app_id}.json", False),
            (f"{app_id}.yaml", True),
            (f"{app_id}.yml", True)]:
        url = f"https://raw.githubusercontent.com/flathub/{app_id}/master/{fname}"
        response = requests.get(url)
        if response.status_code == 404:
            continue
        else:
            break

    response.raise_for_status()

    if is_yaml:
        return yaml.safe_load(response.text)
    else:
        # flatpak-builder supports non-standard comments in the manifest, strip
        # them out. (Ignore the possibility of C comments embedded in strings.)
        no_comments = re.sub(r'/\*.*?\*/', '', response.text, flags=re.DOTALL)
        return json.loads(no_comments)


class FlatpakGenerator(ModuleGenerator):
    # The "api" for a Flatpak is that runs - other modules never depend on it
    include_api_section = False
    # Rationale for packages from the command line
    api_component_rationale = "Application package"

    def __init__(self, pkg):
        super().__init__([pkg])
        self.rpm_branch = dataset_release_branch()

    def _calculate_dependencies(self):
        pkg = self.pkgs[0]
        pool = self._pool

        self.api_srpms = {rpm_name_only(_depchase.get_srpm_for_rpm(pool, pkg))}

        hint_packages = _get_runtime_packages()
        if self.include_flatpak_common:
            hint_packages.update(_get_common_packages())

        all_needed_packages = _depchase.ensure_installable(
            pool, [pkg], hints=hint_packages)

        pkgs = all_needed_packages - hint_packages

        run_srpms = {rpm_name_only(_depchase.get_srpm_for_rpm(pool, dep))
                     for dep in pkgs}
        self.run_srpms = run_srpms - self.api_srpms

        self.module_run_deps = {'flatpak-runtime': [dataset_release_branch()]}
        if self.include_flatpak_common:
            self.module_run_deps['flatpak-common'] = [dataset_release_branch()]
        self.module_run_deps['platform'] = [dataset_release_branch()]

    def _update_module_md(self):
        super()._update_module_md()

        default_profile = Modulemd.Profile(name="default")
        for pkg in self.pkgs:
            default_profile.add_rpm(pkg)

        self.mmd.add_profile(default_profile)

    def _flathub_container_yaml(self, manifest):
        app_id = manifest.get('app-id')
        if app_id is None:
            app_id = manifest['id']
        yml = NoSortMapping({
            'compose': {
                'modules': [self.pkgs[0] + ':master'],
            },
            'flatpak': NoSortMapping({
                'id': app_id,
                'branch': 'stable',
            })
        })

        for key in ['command',
                    'appstream-license',
                    'appstream-compose',
                    'desktop-file-name-prefix',
                    'desktop-file-name-suffix',
                    'rename-appdata-file',
                    'rename-desktop-file',
                    'rename-icon',
                    'copy-icon']:
            if key in manifest:
                yml['flatpak'][key] = manifest[key]

        if 'finish-args' in manifest:
            yml['flatpak']['finish-args'] = LiteralScalar('\n'.join(manifest['finish-args']))

        return yaml.dump(yml, default_flow_style=False, indent=4)

    def _default_container_yaml(self):
        pkg = self.pkgs[0]
        command = pkg

        container_yaml = dedent(f'''\
            compose:
                modules:
                    - {pkg}:master
            flatpak:
                # Derived from the project's domain name
                id: org.example.MyApp
                branch: stable
                # Binary to execute to run the app
                command: {command}
                tags: []
                # Not sandboxed. See 'man flatpak-build-finish'
                finish-args: |
                    --filesystem=host
                    --share=ipc
                    --socket=x11
                    --socket=wayland
                    --socket=session-bus
            ''')

        return container_yaml

    def _write_container_yaml(self, output_fname, flathub_manifest):
        if flathub_manifest:
            container_yaml = self._flathub_container_yaml(flathub_manifest)
        else:
            container_yaml = self._default_container_yaml()

        with open(output_fname, 'w') as f:
            f.write(container_yaml)

        print(f"Generated container specification: {output_fname!r}."
              f" Please edit appropriately.")

    def run(self, output_modulemd, output_containerspec,
            force=False, flathub=None, include_flatpak_common=False):
        flathub_manifest = _load_flathub_manifest(flathub) if flathub else None

        self.include_flatpak_common = include_flatpak_common

        if output_modulemd is None:
            pkg = self.pkgs[0]
            output_modulemd = pkg + '.yaml'
        if output_containerspec is None:
            output_containerspec = 'container.yaml'

        if not force:
            if os.path.exists(output_modulemd):
                raise click.ClickException(f"{output_modulemd} exists. Pass"
                                           f" --force to overwrite.")
            if os.path.exists(output_containerspec):
                raise click.ClickException(f"{output_containerspec} exists."
                                           f" Pass --force to overwrite.")

        super().run(output_modulemd)

        self._write_container_yaml(output_containerspec, flathub_manifest)


def do_flatpak_report(pkgs, quiet=False):
    if not quiet:
        print("Initializing", file=sys.stderr)
    runtime_packages = _get_runtime_packages()
    pool = _depchase.make_pool()

    packages = {}
    flatpaks = {}

    for p in runtime_packages:
        packages[p] = {
            'name': p,
            'runtime': True,
            'used_by': []
        }

    for pkg in pkgs:
        if not quiet:
            print("Calculating deps for", pkg, file=sys.stderr)
        all_needed_packages = _depchase.ensure_installable(
            pool, [pkg], hints=sorted(runtime_packages))
        extra_packages = all_needed_packages - runtime_packages - {pkg}
        used_runtime_packages = all_needed_packages - extra_packages - {pkg}
        flatpaks[pkg] = {
            'runtime': sorted(used_runtime_packages),
            'extra': sorted(extra_packages)
        }
        for p in all_needed_packages:
            if p == pkg:
                continue
            data = packages.get(p, None)
            if data is None:
                data = {
                    'name': p,
                    'runtime': False,
                    'used_by': []
                }
                packages[p] = data
            data['used_by'].append(pkg)

    json.dump({
        'packages': packages,
        'flatpaks': flatpaks
    }, sys.stdout, indent=4)
