# -*- coding: utf-8 -*-
#
# sources.buildsys - functionality for accessing metadata from build systems
#
# Copyright © 2018, 2019 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from fnmatch import fnmatchcase
import logging
from typing import Iterator, Optional, Sequence

import gi
gi.require_version('Modulemd', '1.0')  # noqa: E402
from gi.repository import Modulemd

from .connectors.mbs import MBSConsumer
from .connectors.pagure import PagureConsumer
from .connectors.rest import retry
from .core import Source
from ..config import config, runtime_config
from ..util import parse_dataset_name


log = logging.getLogger(__name__)


class _VerificationError(ValueError):
    pass


class BuildSystemSource(Source):

    source_type = 'buildsys'

    def __init__(self):
        dataset = runtime_config['dataset']
        release_name, arch = parse_dataset_name(dataset)
        self.release = config.releases[release_name]

        self.bs_config = self.release['buildsystems']
        self.pkgs_config = self.bs_config['pkgs']
        self.mbs_config = self.bs_config['mbs']

        self.pkgs = PagureConsumer(base_url=self.pkgs_config['baseurl'],
                                   conn_limit=self.pkgs_config['connection-limit'])
        self.mbs = MBSConsumer(base_url=self.mbs_config['baseurl'],
                               conn_limit=self.mbs_config['connection-limit'])

    @property
    def loop(self):
        return asyncio.get_event_loop()

    @staticmethod
    async def _list_from_async_generators(*generators):
        result = []
        for gen in generators:
            async for x in gen:
                result.append(x)
        return result

    def enumerate_module_names(
        self, module_patterns: Optional[Sequence[str]] = None
    ) -> Iterator[str]:
        log.info("Discovering modules.")

        if module_patterns is None:
            generators = (
                self.pkgs.collect_pages('get_projects', namespace='modules',
                                        fork='false', short='true',
                                        per_page=self.pkgs_config['results-per-page']),
            )
        else:
            generators = (
                self.pkgs.collect_pages('get_projects', namespace='modules',
                                        fork='false', short='true',
                                        pattern=pattern,
                                        per_page=self.pkgs_config['results-per-page'])
                for pattern in module_patterns
            )

        # Keep track of which modules are processed, in case several module
        # patterns have overlapping result sets.
        seen = set()

        for p in self.loop.run_until_complete(
            self._list_from_async_generators(*generators)
        ):
            try:
                mname = p['name']
            except KeyError:
                log.warning(f"Received project without 'name': {p!r}")
            if mname not in seen:
                seen.add(mname)
                log.info(f"Found module: {mname}")
                yield mname

    @staticmethod
    def _vrfy_build_json(build_json, name, stream, build_id=None):
        if build_id is not None:
            build_string = f"'{name}:{stream}[{build_id}]'"
        else:
            build_string = f"'{name}:{stream}'"

        error = False
        if 'id' not in build_json:
            log.warning(f"Received build for {build_string} without 'id' from MBS.")
            error = True
        if 'modulemd' not in build_json:
            log.warning(f"Received build for {build_string} without 'modulemd'"
                        f" from MBS (id: {build_json.get('id') or build_id}")
            error = True

        if error:
            raise _VerificationError()

    async def _sid_and_build(self, sid):
        return sid, await retry(self.mbs.get_module_build,
                                sid=sid, verbose='true',
                                retry_attempts=10)

    async def _get_module_builds_for_stream(
        self,
        name: str,
        stream: str,
        nsvc_field_patterns: Optional[Sequence[str]] = None,
        filter_by_platform: bool = True,
        latest: bool = True
    ) -> Iterator[dict]:
        """Given a module name and stream, this generator coroutine yields builds for the
        various sibling contexts.

        :param name:    name of the module
        :param stream:  name of the stream
        :param nsvc_field_patterns:
                        an optional list of glob patterns to filter names,
                        streams, versions, contexts with
        :param filter_by_platform:
                        whether to only list builds depending on the configured
                        platform module stream
        :param latest:  whether to only list latest builds

        :returns:       generator yielding JSON dictionaries"""

        log.debug(f"_get_module_builds_for_stream(name={name!r}, stream={stream!r},"
                  f" nsvc_field_patterns={nsvc_field_patterns!r}, latest={latest!r})")

        if filter_by_platform:
            platform_mconfig = self.release['platform-module']
            platform_module = platform_mconfig['name']
            platform_stream = platform_mconfig['stream-name']

        build = None

        kwargs = {'name': name, 'stream': stream, 'verbose': 'true',
                  'state': MBSConsumer.STATE_READY}

        if nsvc_field_patterns is not None:
            filtered_patterns = self.filter_nsvc_fields(
                nsvc_field_patterns, (name, stream, None, None)
            )

            if not filtered_patterns:
                raise ValueError(f"NSVC patterns don't match name {name!r}, stream {stream!r}:"
                                 f" {nsvc_field_patterns!r}")

            nsvc_field_patterns = filtered_patterns

        if latest and not nsvc_field_patterns and not filter_by_platform:
            per_page = 1
            kwargs['limit'] = 1
        else:
            per_page = self.mbs_config['results-per-page']

        builds = self.mbs.collect_pages('get_module_builds',
                                        per_page=per_page,
                                        **kwargs)

        async for build in builds:
            try:
                self._vrfy_build_json(build, name, stream)
            except _VerificationError:
                log.warning("--> skipping.")
                continue

            if build is None:
                log.info(f"No build for {name}:{stream} found in MBS.")
                return

            nsvc = (build['name'], build['stream'], build['version'], build['context'])
            nsvc_str = ':'.join(nsvc)

            if nsvc_field_patterns and not self.filter_nsvc_fields(nsvc_field_patterns, nsvc):
                log.debug(f"Ignoring non-matching build: {nsvc_str}")
                continue

            if filter_by_platform:
                buildrequires = build['buildrequires']

                if not buildrequires:
                    log.debug(f"Ignoring build without buildrequires: {nsvc_str}")
                    continue

                if not isinstance(buildrequires, dict):
                    log.debug(f"Ignoring build with buildrequires of wrong"
                              f" type: {nsvc_str} {buildrequires!r}")
                    continue

                build_platform_stream = buildrequires.get(platform_module, {}).get('stream')

                if build_platform_stream != platform_stream:
                    log.debug(f"Ignoring build depending on wrong platform stream:"
                              f" {nsvc_str} -> {platform_module}:{build_platform_stream}")
                    continue

            log.info(f"Found build: {nsvc_str}")

            yield build

            if latest:
                # If we're only interested in the latest successful build,
                # gather sibling builds and break out of loop.
                for sid, sbuild in await asyncio.gather(*(
                    self._sid_and_build(sid)
                    for sid in build.get('siblings', ())
                )):
                    try:
                        self._vrfy_build_json(sbuild, name, stream, build_id=sid)
                    except _VerificationError:
                        log.warning("--> skipping.")
                        continue

                    nsvc = (sbuild['name'], sbuild['stream'],
                            sbuild['version'], sbuild['context'])
                    nsvc_str = ':'.join(nsvc)

                    if nsvc_field_patterns and not self.filter_nsvc_fields(
                        nsvc_field_patterns, nsvc
                    ):
                        log.debug(f"Ignoring non-matching build: {nsvc_str}")
                        continue

                    if filter_by_platform:
                        buildrequires = sbuild['buildrequires']

                        if not buildrequires:
                            log.debug(f"Ignoring build without buildrequires: {nsvc_str}")
                            continue

                        if not isinstance(buildrequires, dict):
                            log.debug(f"Ignoring build with buildrequires of wrong"
                                      f" type: {nsvc_str} {buildrequires!r}")
                            continue

                        build_platform_stream = buildrequires.get(platform_module, {}).get('stream')

                        if build_platform_stream != platform_stream:
                            log.debug(f"Ignoring build depending on wrong platform stream:"
                                      f" {nsvc_str} -> {platform_module}:{platform_stream}")
                            continue

                    log.info(f"Found build: {nsvc_str}")

                    yield sbuild

                break

    async def _get_module_builds_for_module(
        self,
        name: str,
        nsvc_field_patterns: Optional[Sequence[str]] = None,
        latest: bool = True
    ) -> Iterator[dict]:
        """Given a module name, this generator yields builds for all streams
        and their contexts.

        :param name:          the name of the module
        :param nsvc_field_patterns:
                              an optional list of glob patterns to filter
                              names, streams, versions, contexts with
        :param latest:        whether to return only latest builds for a
                              stream

        :returns:             a generator yielding dictionaries describing
                              the builds"""

        log.debug(f"_get_module_builds_for_module(name={name!r})")

        try:
            res = await retry(self.pkgs.get_project_branches,
                              namespace='modules', project=name,
                              retry_attempts=5)
            streams = res['branches']
        except KeyError:
            streams = None
        else:
            if nsvc_field_patterns is not None:
                filtered_streams = []
                for s in streams:
                    for p in nsvc_field_patterns:
                        np, sp = (f or '*' for f in p[:2])
                        if fnmatchcase(name, np) and fnmatchcase(s, sp):
                            filtered_streams.append(s)
                streams = filtered_streams

        if not streams:
            log.warning(f"No matching stream branches found for '{name}'.")
            return

        for build in await self._list_from_async_generators(*(
                    self._get_module_builds_for_stream(name,
                                                       stream,
                                                       nsvc_field_patterns,
                                                       latest=latest)
                    for stream in streams
        )):
            yield build

    async def _get_module_builds_from_mbs(self):
        """Asynchronous generator that yields all builds known to MBS."""

        async for build in self.mbs.collect_pages(
                'get_module_builds', verbose='true',
                per_page=self.mbs_config['results-per-page']):
            yield build

    def enumerate_module_builds(
        self,
        nsvc_patterns: Optional[Sequence[str]] = None,
        latest: bool = True,
    ) -> Iterator[Modulemd.Module]:
        """Given a list of module NSVC patterns (defaulting to "everything"),
        this generator yields builds for all streams and contexts for these
        modules.

        :param nsvc_patterns: a list of partial or complete module NSVCs
                              which are globs to filter with (optional)
        :param latest:        whether to list only latest builds

        :returns:             a generator yielding Modulemd.Module objects"""

        log.info("Querying module builds.")
        log.debug(f"enumerate_module_builds({nsvc_patterns!r}, {latest!r})")

        if nsvc_patterns:
            nsvc_patterns_fields = [self.split_nsvc_fields(pat)
                                    for pat in nsvc_patterns]

            mname_patterns = {pat[0] for pat in nsvc_patterns_fields}
        else:
            mname_patterns = None

        module_names = list(self.enumerate_module_names(mname_patterns))

        log.debug(f"\tmodule_names = {module_names!r}")

        if nsvc_patterns:
            module_names_fields = {}
            for mname in module_names:
                flist = module_names_fields[mname] = []
                for pf in nsvc_patterns_fields:
                    np = pf[0]
                    if fnmatchcase(mname, np):
                        flist.append(pf)

            log.debug(f"\tmodule_names_fields = {module_names_fields!r}")
        else:
            module_names_fields = {}

        builds = self.loop.run_until_complete(self._list_from_async_generators(*(
            self._get_module_builds_for_module(
                name,
                nsvc_field_patterns=module_names_fields.get(name),
                latest=latest
            )
            for name in module_names
        )))

        for build in builds:
            module = Modulemd.Module.new_from_string(build['modulemd'])
            module.buildinfo = build
            yield module

    def enumerate_module_defaults(
        self, module_patterns: Optional[Sequence[str]] = None
    ) -> Iterator[Modulemd.Defaults]:
        """Given a list of module NSVC patterns (defaulting to "everything"),
        this generator yields builds for all streams and contexts for these
        modules.

        Because the build systems involved (MBS, Koji, dist-git/Pagure) know
        nothing about defaults, this method will actually yield nothing.

        :param module_patterns: a list of module name patterns which are globs
                                to filter with (optional)

        :returns:               a generator yielding Modulemd.Default objects
        """

        return
        yield
