# -*- coding: utf-8 -*-
#
# sources.connectors.rest - common functionality to access REST services
#
# Copyright © 2018, 2019 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author:
# Nils Philippsen <nils@redhat.com>

import asyncio
import inspect
import json
import logging
import math
from urllib.parse import urljoin

import aiohttp


log = logging.getLogger(__name__)

_unset = object()


def _json_access_tree(tree, path):
    node = tree
    if path:
        for path_elem in path.split('.'):
            node = node[path_elem]
    return node


async def retry(coroutine,
                retry_attempts=2,
                sleep=1,
                max_sleep=30,
                sleep_growth=1.5,
                catch_exceptions=(aiohttp.ClientConnectorError,
                                  aiohttp.ClientResponseError,
                                  aiohttp.ServerDisconnectedError,
                                  asyncio.TimeoutError,
                                  json.JSONDecodeError),
                *args, **kwargs):
    """Retry a coroutine until it succeeds (or the number of attempts is
    exceeded).

    :param coroutine:           an async coroutine
    :param retry_attempts:      how many attempts to undertake
    :param catch_exceptions:    which exceptions cause retries

    :param args:                positional parameters passed to the coroutine
    :param kwargs:              keyword parameters passed to the coroutine"""

    # some smoke tests
    if retry_attempts <= 0:
        raise ValueError("'retry_attempts' must be >= 1")
    if not len(catch_exceptions):
        raise ValueError("'catch_exceptions' must not be empty")

    for attempt in range(retry_attempts):
        if attempt:
            log.debug(f">>> {coroutine}(*args={args!r}, **kwargs={kwargs!r}) retry {attempt}")
        try:
            retval = await coroutine(*args, **kwargs)
        except catch_exceptions as e:
            log.debug(f">>>>>> {e!r}")
            if attempt >= retry_attempts - 1:
                log.error(f">>> failed: {coroutine}(*args={args!r}, **kwargs={kwargs!r})")
                log.error(f">>> attempt {attempt + 1} of {retry_attempts}\n>>> {e!r}")
                raise
            # sleep a while
            log.debug(f">>>>>> sleep {sleep} seconds, then retry")
            await asyncio.sleep(sleep)
            sleep = min(math.ceil(sleep * sleep_growth), max_sleep)
            # try again until retry_attempts is exceeded
        else:
            break

    return retval


def paginate(**kwargs):
    """
    Decorate callables with pagination parameters.

    :param page_arg:        the query parameter to specify the desired page
    :param page_elem:       the JSON element which contains the current page
                            number, the value of :page_arg: is assumed if
                            unspecified
    :param pages_elem:      the JSON element which contains the total number of
                            pages
    :param data_elem:       the JSON element which contains the desired data
    :param retry_attempts:  how often a page should be attempted to be
                            retrieved on errors before failing
    """
    def _paginate_decorate(func):
        func.pagination = dict(**kwargs)
        return func
    return _paginate_decorate


class PaginateMixin:

    """A mixin to augment REST API consumer classes with a collect_pages()
    method to transparently 'flip' through paginated API output, and matching
    defaults."""

    pagination = {
        # the query parameter to specify the desired page
        'page_arg': 'page',
        # the JSON element which contains the current page number, the
        # value of page_arg is assumed if unspecified
        'page_elem': None,
        # the JSON element which contains the total number of pages
        'pages_elem': 'pages',
        # the JSON element which contains the desired data
        'data_elem': 'data',
        # how often a page should be attempted to be retrieved on errors before
        # failing
        'retry_attempts': 5,
    }

    @staticmethod
    def _merge_dict_shallow(dct1, dct2):
        """Performs a shallow merge (one level) of dct1 and dct2. Dictionary
        dct1 is the 'source' and dct2 the 'target', i.e. dct1 is unchanged,
        dct2 is updated with elements from dct1 which don't exist by the same
        name."""

        for k, v in dct1.items():
            if k not in dct2:
                dct2[k] = v

    def __init_subclass__(cls):
        if not hasattr(cls, 'pagination'):
            cls.pagination = {}

        for base in cls.__bases__:
            if hasattr(base, 'pagination'):
                cls._merge_dict_shallow(base.pagination, cls.pagination)

    async def collect_pages(self, func, pagination=None, limit=None, *args, **kwargs):
        """A generator to 'flip' through paginated JSON API output and gather
        the results. Yields one element at a time.

        :param func:        A callable, or name of a method.
        :param pagination:  Optional pagination parameters for this call.
        :param limit:       Optional limit of pages to return.
        :param args:        Positional arguments passed on to the callable.
        :param kwargs:      Keyword arguments passed on to the callable."""

        if isinstance(func, str):
            func = getattr(self, func)

        if pagination is None:
            pagination = {}
        else:
            # don't overwrite the parameter below
            pagination = pagination.copy()

        if hasattr(func, 'pagination'):
            self._merge_dict_shallow(func.pagination, pagination)

        self._merge_dict_shallow(self.pagination, pagination)

        page_arg = pagination['page_arg']
        data_elem = pagination['data_elem']
        pages_elem = pagination['pages_elem']
        retry_attempts = pagination['retry_attempts']

        # For pagination, we have to add the page argument in the right place, e.g.
        # if a decorated callable has it as a parameter with a default value it
        # can be set in args and needs to be replaced there rather than adding it
        # to kwargs to avoid exceptions like:
        #
        # TypeError: ... got multiple values for argument 'page'

        # Find out if the page arg is set as a a positional parameter. If yes,
        # remember its position in page_arg_idx and convert args to a list so it
        # can be replaced on the fly.
        sig = inspect.signature(func)
        try:
            page_arg_idx = list(sig.parameters)[:len(args)].index(page_arg)
        except ValueError:
            page_arg_idx = None
        else:
            args = list(args)

        # we have to start with something, one page is something
        current_ceiling = 1
        pages_retrieved = 0

        result_pages = []

        while pages_retrieved < current_ceiling:
            retr_coros = []

            # schedule one concurrent request per page, limited by aiohttp's
            # session connector
            for page in range(pages_retrieved + 1, current_ceiling + 1):
                if page_arg_idx is not None:
                    # page arg is set as a positional parameter, replace it
                    args[page_arg_idx] = page
                else:
                    kwargs[page_arg] = page

                retr_coros.append(retry(
                    func, retry_attempts=retry_attempts, *args, **kwargs))

            partial_result_pages = await asyncio.gather(*retr_coros)
            if not len(partial_result_pages):
                break

            result_pages.extend(partial_result_pages)
            pages_retrieved = len(result_pages)
            current_ceiling = _json_access_tree(result_pages[-1], pages_elem)
            if limit is not None:
                current_ceiling = min(limit, current_ceiling)

        for page_content in result_pages:
            for elem in page_content[data_elem]:
                yield elem


class Consumer(PaginateMixin):
    """A lightweight, asynchronous REST API consumer."""

    def __init__(self, base_url, conn_limit=10):
        self.base_url = base_url

        self.connector = aiohttp.TCPConnector(limit=conn_limit)
        self.session = aiohttp.ClientSession(connector=self.connector,
                                             raise_for_status=True)

    def __del__(self):
        # closing the session will close the connector, too
        asyncio.get_event_loop().run_until_complete(self.session.close())

    def _endpoint(self, path):
        """Build a full URL from the base URL of a service and the path to the
        endpoint."""
        return urljoin(self.base_url, path)

    async def get(self, path, params=None, **kwargs):
        """Combine base URL and path and execute a GET request.

        :param path: URL relative to the base"""

        return await self.session.get(self._endpoint(path), params=params, **kwargs)

    async def get_as_json(self, path, params=None, **kwargs):
        """Combine base URL and path, execute a GET request and parse it as a
        JSON object.

        :param path: URL relative to the base"""
        resp = await self.get(path, params=params, **kwargs)

        return await resp.json()
